@@ -14,6 +14,8 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
+#define __STDC_LIMIT_MACROS 1
+
 #include <assert.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -89,6 +91,7 @@ SRATOM_API
 void
 sratom_free(Sratom* sratom)
 {
+	serd_node_free(&sratom->base_uri);
 	free(sratom);
 }
 
@@ -183,7 +186,11 @@ start_object(Sratom*         sratom,
 	if (subject && predicate) {
 		sratom->write_statement(sratom->handle, *flags|SERD_ANON_O_BEGIN, NULL,
 		                        subject, predicate, node, NULL, NULL);
+		// Start abbreviating object properties
 		*flags |= SERD_ANON_CONT;
+
+		// Object is in a list, stop list abbreviating if necessary
+		*flags &= ~SERD_LIST_CONT;
 	}
 	if (type) {
 		SerdNode p = serd_node_from_string(SERD_URI, NS_RDF "type");
@@ -228,6 +235,7 @@ sratom_write(Sratom*         sratom,
 	} else if (type_urid == sratom->forge.Chunk) {
 		datatype = serd_node_from_string(SERD_URI, NS_XSD "base64Binary");
 		object   = serd_node_new_blob(body, size, true);
+		new_node = true;
 	} else if (type_urid == sratom->forge.Literal) {
 		const LV2_Atom_Literal_Body* lit = (const LV2_Atom_Literal_Body*)body;
 		const uint8_t*         str = USTR(lit + 1);
@@ -254,7 +262,7 @@ sratom_write(Sratom*         sratom,
 		const uint8_t* str = USTR(body);
 		if (path_is_absolute((const char*)str)) {
 			new_node = true;
-			object   = serd_node_new_file_uri(str, NULL, NULL, false);
+			object   = serd_node_new_file_uri(str, NULL, NULL, true);
 		} else {
 			SerdURI base_uri = SERD_URI_NULL;
 			if (!sratom->base_uri.buf ||
@@ -268,7 +276,7 @@ sratom_write(Sratom*         sratom,
 					serd_uri_parse(sratom->base_uri.buf, &base_uri);
 				}
 				new_node = true;
-				SerdNode rel = serd_node_new_file_uri(str, NULL, NULL, false);
+				SerdNode rel = serd_node_new_file_uri(str, NULL, NULL, true);
 				object = serd_node_new_uri_from_node(&rel, &base_uri, NULL);
 				serd_node_free(&rel);
 			}
@@ -416,6 +424,14 @@ sratom_write(Sratom*         sratom,
 	}
 
 	if (object.buf) {
+		SerdNode def_s = serd_node_from_string(SERD_BLANK, USTR("atom"));
+		SerdNode def_p = serd_node_from_string(SERD_URI, USTR(NS_RDF "value"));
+		if (!subject) {
+			subject = &def_s;
+		}
+		if (!predicate) {
+			predicate = &def_p;
+		}
 		sratom->write_statement(sratom->handle, flags, NULL,
 		                        subject, predicate, &object, &datatype, &language);
 	}
@@ -474,26 +490,10 @@ sratom_to_turtle(Sratom*         sratom,
 
 	serd_writer_free(writer);
 	serd_env_free(env);
+	serd_node_free(&base);
 	return (char*)serd_chunk_sink_finish(&str);
 }
 
-static const SordNode*
-get_object(SordModel*      model,
-           const SordNode* subject,
-           const SordNode* predicate)
-{
-	const SordNode* object = NULL;
-	SordQuad        q      = { subject, predicate, 0, 0 };
-	SordIter*       i      = sord_find(model, q);
-	if (!sord_iter_end(i)) {
-		SordQuad quad;
-		sord_iter_get(i, quad);
-		object = quad[SORD_OBJECT];
-	}
-	sord_iter_free(i);
-	return object;
-}
-
 static void
 read_list_value(Sratom*         sratom,
                 LV2_Atom_Forge* forge,
@@ -502,12 +502,14 @@ read_list_value(Sratom*         sratom,
                 const SordNode* node,
                 ReadMode        mode)
 {
-	const SordNode* first = get_object(model, node, sratom->nodes.rdf_first);
-	const SordNode* rest  = get_object(model, node, sratom->nodes.rdf_rest);
-	if (first && rest) {
-		read_node(sratom, forge, world, model, first, mode);
-		read_list_value(sratom, forge, world, model, rest, mode);
+	SordNode* fst = sord_get(model, node, sratom->nodes.rdf_first, NULL, NULL);
+	SordNode* rst = sord_get(model, node, sratom->nodes.rdf_rest, NULL, NULL);
+	if (fst && rst) {
+		read_node(sratom, forge, world, model, fst, mode);
+		read_list_value(sratom, forge, world, model, rst, mode);
 	}
+	sord_node_free(world, rst);
+	sord_node_free(world, fst);
 }
 
 static void
@@ -636,8 +638,10 @@ read_node(Sratom*         sratom,
 			lv2_atom_forge_urid(forge, map->map(map->handle, str));
 		}
 	} else {
-		const SordNode* type  = get_object(model, node, sratom->nodes.rdf_type);
-		const SordNode* value = get_object(model, node, sratom->nodes.rdf_value);
+		SordNode* type = sord_get(
+			model, node, sratom->nodes.rdf_type, NULL, NULL);
+		SordNode* value = sord_get(
+			model, node, sratom->nodes.rdf_value, NULL, NULL);
 
 		const uint8_t* type_uri  = NULL;
 		uint32_t       type_urid = 0;
@@ -648,13 +652,14 @@ read_node(Sratom*         sratom,
 
 		LV2_Atom_Forge_Frame frame = { 0, 0 };
 		if (mode == MODE_SEQUENCE) {
-			const SordNode* frame_time = get_object(
-				model, node, sratom->nodes.atom_frameTime);
+			SordNode* frame_time = sord_get(
+				model, node, sratom->nodes.atom_frameTime, NULL, NULL);
 			const char* frame_time_str = frame_time
 				? (const char*)sord_node_get_string(frame_time)
 				: "";
 			lv2_atom_forge_frame_time(forge, serd_strtod(frame_time_str, NULL));
 			read_node(sratom, forge, world, model, value, MODE_BODY);
+			sord_node_free(world, frame_time);
 		} else if (type_urid == sratom->forge.Tuple) {
 			lv2_atom_forge_tuple(forge, &frame);
 			read_list_value(sratom, forge, world, model, value, MODE_BODY);
@@ -662,8 +667,8 @@ read_node(Sratom*         sratom,
 			lv2_atom_forge_sequence_head(forge, &frame, 0);
 			read_list_value(sratom, forge, world, model, value, MODE_SEQUENCE);
 		} else if (type_urid == sratom->forge.Vector) {
-			const SordNode* child_type_node = get_object(
-				model, node, sratom->nodes.atom_childType);
+			SordNode* child_type_node = sord_get(
+				model, node, sratom->nodes.atom_childType, NULL, NULL);
 			uint32_t child_type = map->map(
 				map->handle, (const char*)sord_node_get_string(child_type_node));
 			uint32_t child_size = atom_size(sratom, child_type);
@@ -671,6 +676,7 @@ read_node(Sratom*         sratom,
 				lv2_atom_forge_vector_head(forge, &frame, child_size, child_type);
 				read_list_value(sratom, forge, world, model, value, MODE_BODY);
 			}
+			sord_node_free(world, child_type_node);
 		} else if (value && sord_node_equals(sord_node_get_datatype(value),
 		                                     sratom->nodes.xsd_base64Binary)) {
 			size_t         vlen = 0;
@@ -691,6 +697,8 @@ read_node(Sratom*         sratom,
 		if (frame.ref) {
 			lv2_atom_forge_pop(forge, &frame);
 		}
+		sord_node_free(world, value);
+		sord_node_free(world, type);
 	}
 }
 
@@ -764,16 +772,13 @@ sratom_from_turtle(Sratom*         sratom,
 			&sratom->forge, sratom_forge_sink, sratom_forge_deref, &out);
 		if (subject && predicate) {
 			SordNode* p = sord_node_from_serd_node(world, env, predicate, 0, 0);
-			SordQuad  q = { s, p, 0, 0 };
-			SordIter* i = sord_find(model, q);
-			if (!sord_iter_end(i)) {
-				SordQuad result;
-				sord_iter_get(i, result);
-				sratom_read(sratom, &sratom->forge, world, model, result[SORD_OBJECT]);
+			SordNode* o = sord_get(model, s, p, NULL, NULL);
+			if (o) {
+				sratom_read(sratom, &sratom->forge, world, model, o);
+				sord_node_free(world, o);
 			} else {
 				fprintf(stderr, "Failed to find node\n");
 			}
-			sord_iter_free(i);
 		} else {
 			sratom_read(sratom, &sratom->forge, world, model, s);
 		}
@@ -785,6 +790,7 @@ sratom_from_turtle(Sratom*         sratom,
 	serd_env_free(env);
 	sord_free(model);
 	sord_world_free(world);
+	serd_node_free(&base);
 
 	return (LV2_Atom*)out.buf;
 }
